"""Generate and serve exploratory applications."""
from pathlib import Path
from typing import Callable
import pandas as pd
import panel as pn
from panel.template import BootstrapTemplate

from nwm_explorer.mappings import (EVALUATIONS, DOMAIN_STRINGS,
    DOMAIN_CONFIGURATION_MAPPING, Domain, Configuration, LEAD_TIME_VALUES,
    CONFIDENCE_STRINGS, Confidence, METRIC_STRINGS, Metric)
from nwm_explorer.readers import MetricReader, DashboardState

pn.extension("plotly")

class FilteringWidgets:
    def __init__(self):
        # Filtering options
        self.callbacks: list[Callable] = []
        self.evaluation_filter = pn.widgets.Select(
            name="Evaluation",
            options=list(EVALUATIONS.keys())
        )
        self.domain_filter = pn.widgets.Select(
            name="Model Domain",
            options=list(DOMAIN_STRINGS.keys())
        )
        self.configuration_filter = pn.widgets.Select(
            name="Model Configuration",
            options=list(
            DOMAIN_CONFIGURATION_MAPPING[self.current_domain].keys()
            ))
        self.threshold_filter = pn.widgets.Select(
            name="Streamflow Threshold (≥)",
            options=[
                "100% AEP-USGS (All data)"
            ]
        )
        self.metric_filter = pn.widgets.Select(
            name="Evaluation Metric",
            options=list(METRIC_STRINGS.keys())
        )
        self.confidence_filter = pn.widgets.Select(
            name="Confidence Estimate (95%)",
            options=list(CONFIDENCE_STRINGS.keys())
        )
        if self.current_configuration in LEAD_TIME_VALUES:
            options = LEAD_TIME_VALUES[self.current_configuration]
        else:
            options = [0]
        self.lead_time_filter = pn.Row(pn.widgets.DiscretePlayer(
            name="Minimum lead time (hours)",
            options=options,
            show_loop_controls=False,
            visible_buttons=["previous", "next"],
            width=300
            ))

        def handle_domain_change(domain):
            if domain is None:
                return
            self.update_configurations()
        pn.bind(handle_domain_change, self.domain_filter, watch=True)

        def handle_configuration_change(domain):
            if domain is None:
                return
            self.update_lead_times()
        pn.bind(handle_configuration_change, self.configuration_filter,
            watch=True)
        
    @property
    def current_start_date(self) -> pd.Timestamp:
        return EVALUATIONS[self.evaluation_filter.value][0]
        
    @property
    def current_end_date(self) -> pd.Timestamp:
        return EVALUATIONS[self.evaluation_filter.value][1]

    @property
    def current_domain(self) -> Domain:
        return DOMAIN_STRINGS[self.domain_filter.value]

    @property
    def current_configuration(self) -> Configuration:
        return DOMAIN_CONFIGURATION_MAPPING[self.current_domain][self.configuration_filter.value]
    
    @property
    def current_lead_time(self) -> int:
        return self.lead_time_filter[0].value

    @property
    def current_confidence(self) -> Confidence:
        return CONFIDENCE_STRINGS[self.confidence_filter.value]

    @property
    def current_metric(self) -> Metric:
        return METRIC_STRINGS[self.metric_filter.value]

    def update_configurations(self) -> None:
        """Set configuration options"""
        self.configuration_filter.options = list(
            DOMAIN_CONFIGURATION_MAPPING[self.current_domain].keys())

    def update_lead_times(self) -> None:
        """Set lead time options"""
        c = self.current_configuration
        if c in LEAD_TIME_VALUES:
            options = LEAD_TIME_VALUES[c]
        else:
            options = [0]
        
        self.lead_time_filter.objects = [
            pn.widgets.DiscretePlayer(
                name="Minimum lead time (hours)",
                options=options,
                show_loop_controls=False,
                visible_buttons=["previous", "next"],
                width=300
                )
        ]
        for func in self.callbacks:
            pn.bind(func, self.lead_time_filter[0], watch=True)

    @property
    def state(self) -> DashboardState:
        """Current widget states."""
        return DashboardState(
            start_date=self.current_start_date,
            end_date=self.current_end_date,
            domain=self.current_domain,
            configuration=self.current_configuration,
            threshold=self.threshold_filter.value,
            metric=self.current_metric,
            confidence=self.current_confidence,
            lead_time=self.current_lead_time
        )

    @property
    def layout(self) -> pn.Column:
        return pn.Column(
            self.evaluation_filter,
            self.domain_filter,
            self.configuration_filter,
            self.threshold_filter,
            self.metric_filter,
            self.confidence_filter,
            self.lead_time_filter
        )
    
    def register_callback(self, func: Callable) -> None:
        """Register callback function."""
        pn.bind(func, self.evaluation_filter, watch=True)
        pn.bind(func, self.domain_filter, watch=True)
        pn.bind(func, self.configuration_filter, watch=True)
        pn.bind(func, self.threshold_filter, watch=True)
        pn.bind(func, self.metric_filter, watch=True)
        pn.bind(func, self.confidence_filter, watch=True)
        pn.bind(func, self.lead_time_filter[0], watch=True)
        self.callbacks.append(func)

class Dashboard:
    """Build a dashboard for exploring National Water Model output."""
    def __init__(self, root: Path, title: str):
        # Data reader
        self.reader = MetricReader(root)
    
        # Get widgets
        self.filter_widgets = FilteringWidgets()

        # Layout filtering options
        self.filter_card = pn.Card(
            self.filter_widgets.layout,
            title="Filters",
            collapsible=False
            )
        
        # Setup map
        self.site_map = pn.pane.Plotly(
            self.reader.get_plotly_patch(self.state))
        self.map_card = pn.Card(
            self.site_map,
            collapsible=False,
            hide_header=True
            )
        
        def update_map(event):
            if event is None:
                return
            self.site_map.object = self.reader.get_plotly_patch(self.state)
        self.filter_widgets.register_callback(update_map)

        # Layout cards
        layout = pn.Row(self.filter_card, self.map_card)
        self.template = BootstrapTemplate(title=title)
        self.template.main.append(layout)

    @property
    def state(self) -> DashboardState:
        """Current dashboard state."""
        return self.filter_widgets.state

def generate_dashboard(
        root: Path,
        title: str
        ) -> BootstrapTemplate:
    return Dashboard(root, title).template

def generate_dashboard_closure(
        root: Path,
        title: str
        ) -> BootstrapTemplate:
    def closure():
        return generate_dashboard(root, title)
    return closure

def serve_dashboard(
        root: Path,
        title: str
        ) -> None:
    # Slugify title
    slug = title.lower().replace(" ", "-")

    # Serve
    endpoints = {
        slug: generate_dashboard_closure(root, title)
    }
    pn.serve(endpoints)



"""Read-only methods."""
from pathlib import Path
from dataclasses import dataclass
from typing import TypedDict
import numpy as np
import pandas as pd
import polars as pl
import plotly.graph_objects as go

from nwm_explorer.mappings import (Domain, Configuration, Metric, Confidence,
    NWM_URL_BUILDERS, FileType, Variable, Units)
from nwm_explorer.urls import generate_reference_dates
from nwm_explorer.data import generate_filepath
from nwm_explorer.logger import get_logger

def read_NWM_output(
        root: Path,
        start_date: pd.Timestamp,
        end_date: pd.Timestamp,
        low_memory: bool = False
) -> dict[tuple[Domain, Configuration], pl.LazyFrame]:
    """
    Find and lazily load NWM output.

    Parameters
    ----------
    root: Path, required
        Root directory to save downloaded and processed files.
    start_date: pd.Timestamp
        First date to start retrieving data.
    end_date: pd.Timestamp
        Last date to retrieve data.
    low_memory: bool, default False
        Reduce memory pressure at the expense of performance.
    
    Returns
    -------
    dict[tuple[Domain, Configuration], pl.LazyFrame]
    """
    logger = get_logger("nwm_explorer.readers.read_NWM_output")
    reference_dates = generate_reference_dates(start_date, end_date)

    # Download and process model output
    model_output = {}
    for (domain, configuration), _ in NWM_URL_BUILDERS.items():
        day_files = []
        for rd in reference_dates:
            # Check for file existence
            parquet_file = generate_filepath(
                root, FileType.parquet, configuration, Variable.streamflow,
                Units.cubic_feet_per_second, rd
            )
            logger.info(f"Building {parquet_file}")
            if parquet_file.exists():
                logger.info(f"Found existing {parquet_file}")
                day_files.append(parquet_file)
                continue
            
        # Check for at least one file
        if not day_files:
            logger.info(f"Found no data for {domain} {configuration}")
            continue
        
        # Merge files
        logger.info(f"Merging parquet files")
        model_output[(domain, configuration)] = pl.scan_parquet(day_files,
            low_memory=low_memory)
    return model_output

def read_USGS_observations(
        root: Path,
        start_date: pd.Timestamp,
        end_date: pd.Timestamp
) -> dict[Domain, pl.LazyFrame]:
    """
    Download and process USGS observations.

    Parameters
    ----------
    root: Path, required
        Root directory to save downloaded and processed files.
    start_date: pd.Timestamp
        First date to start retrieving data.
    end_date: pd.Timestamp
        Last date to retrieve data.
    
    Returns
    -------
    dict[Domain, pl.LazyFrame]
    """
    logger = get_logger("nwm_explorer.pipelines.load_USGS_observations")
    # Download and process model output
    observations = {}
    s = start_date.strftime("%Y%m%dT%H")
    e = end_date.strftime("%Y%m%dT%H")
    for domain in list(Domain):
        # Check for file existence
        parquet_directory = root / FileType.parquet / Configuration.usgs
        parquet_file = parquet_directory / f"{domain}_{s}_{e}.parquet"
        logger.info(f"Building {parquet_file}")
        if parquet_file.exists():
            logger.info(f"Found existing {parquet_file}")
            observations[domain] = pl.scan_parquet(parquet_file)
    return observations

def scan_date_range(
        predictions: dict[tuple[Domain, Configuration], pl.LazyFrame]
) -> tuple[pd.Timestamp, pd.Timestamp]:
    first = None
    last = None
    for (domain, _), data in predictions.items():
        start = data.select("value_time").min().collect().item(0, 0)
        end = data.select("value_time").max().collect().item(0, 0)
        if first is None:
            first = start
            last = end
        else:
            first = min(first, start)
            last = max(last, end)
    return pd.Timestamp(first), pd.Timestamp(last)

def read_pairs(
        root: Path,
        start_date: pd.Timestamp,
        end_date: pd.Timestamp
) -> dict[tuple[Domain, Configuration], pl.LazyFrame]:
    """
    Find and lazily load paired data.

    Parameters
    ----------
    root: Path, required
        Root directory to save downloaded and processed files.
    start_date: pd.Timestamp
        First date to start retrieving data.
    end_date: pd.Timestamp
        Last date to retrieve data.
    
    Returns
    -------
    dict[tuple[Domain, Configuration], pl.LazyFrame]
    """
    logger = get_logger("nwm_explorer.readers.read_pairs")
    # Process model output
    logger.info("Scanning pairs")
    reference_dates = generate_reference_dates(start_date, end_date)
    pairs = {}
    for (domain, configuration), _ in NWM_URL_BUILDERS.items():
        day_files = []
        for rd in reference_dates:
            # Check for file existence
            parquet_file = generate_filepath(
                root, FileType.parquet, configuration, Variable.streamflow_pairs,
                Units.cubic_feet_per_second, rd
            )
            if parquet_file.exists():
                logger.info(f"Found existing file: {parquet_file}")
                day_files.append(parquet_file)
                continue
            
        # Check for at least one file
        if not day_files:
            logger.info(f"Found no data for {domain} {configuration}")
            continue
        
        # Merge files
        logger.info(f"Merging parquet files")
        pairs[(domain, configuration)] = pl.scan_parquet(day_files)
    return pairs

def read_metrics(
        root: Path,
        start_date: pd.Timestamp,
        end_date: pd.Timestamp
) -> dict[tuple[Domain, Configuration], pl.LazyFrame]:
    """
    Scan and lazily load metrics.

    Parameters
    ----------
    root: Path, required
        Root directory to save downloaded and processed files.
    start_date: pd.Timestamp
        First date to start retrieving data.
    end_date: pd.Timestamp
        Last date to retrieve data.
    
    Returns
    -------
    dict[tuple[Domain, Configuration], pl.LazyFrame]
    """
    logger = get_logger("nwm_explorer.pipelines.load_metrics")
    logger.info("Reading pairs")
    pairs = read_pairs(
        root=root,
        start_date=start_date,
        end_date=end_date
        )

    results = {}
    s = start_date.strftime("%Y%m%dT%H")
    e = end_date.strftime("%Y%m%dT%H")
    for (domain, configuration), paired in pairs.items():
        # Check for file existence
        parquet_directory = root / FileType.parquet / "metrics"
        parquet_file = parquet_directory / f"{domain}_{configuration}_{s}_{e}.parquet"
        logger.info(f"Building {parquet_file}")
        if parquet_file.exists():
            logger.info(f"Found existing {parquet_file}")
            results[(domain, configuration)] = pl.scan_parquet(parquet_file)
    return results

class FigurePatch(TypedDict):
    """
    A plotly figure patch.

    Keys
    ----
    data: list[go.Trace]
        A list of plotly traces.
    layout: go.Layout
        Plotly layout.
    """
    data: list[go.Trace]
    layout: go.Layout

@dataclass
class DashboardState:
    """Dashboard state variables."""
    start_date: pd.Timestamp
    end_date: pd.Timestamp
    domain: Domain
    configuration: Configuration
    threshold: str
    metric: Metric
    confidence: Confidence
    lead_time: int

@dataclass
class MetricReader:
    """Intermediate metric reader to query and return data to dashboards."""
    root: Path

    def query(self, state: DashboardState) -> int:
        """Return data matching dashboard state."""
        return str(state).replace(",", "<br>")
    
    def get_plotly_patch(self, state: DashboardState) -> FigurePatch:
        """Return map of sites matching dashboard state."""
        xx = np.linspace(-3.5, 3.5, 100)
        yy = np.linspace(-3.5, 3.5, 100)
        x, y = np.meshgrid(xx, yy)
        z = np.exp(-((x - 1) ** 2) - y**2) - (x**3 + y**4 - x / 5) * np.exp(-(x**2 + y**2))

        surface = go.Surface(z=z)
        layout = go.Layout(
            title=str(state.domain),
            autosize=False,
            width=500,
            height=500,
            margin=dict(t=50, b=50, r=50, l=50)
        )

        return dict(data=[surface], layout=layout)



"""Various methods used to compute model evaluation metrics."""
from typing import Callable
import pandas as pd
import polars as pl
import numpy as np
import numpy.typing as npt
from arch.bootstrap import StationaryBootstrap, optimal_block_length
from numba import float64, guvectorize

@guvectorize([(float64[:], float64[:], float64[:])], "(n),(n)->()")
def nash_sutcliffe_efficiency(
    y_true: npt.NDArray[np.float64],
    y_pred: npt.NDArray[np.float64],
    result: npt.NDArray[np.float64]
    ) -> None:
    """
    Numba and polars compatible implementation of signed mean relative bias.
    Also called, mean relative error or fractional bias.
        
    Parameters
    ----------
    y_true: NDArray[np.float64], required
        Ground truth (correct) target values, also called observations, measurements, or observed values.
    y_pred: NDArray[np.float64], required
        Estimated target values, also called simulations or modeled values.
    result: NDArray[np.float64], required
        Stores scalar result.
        
    Returns
    -------
    None
    """
    variance = np.sum((y_true - np.mean(y_true)) ** 2.0)
    if variance == 0:
        result[0] = np.nan
        return
    result[0] = 1.0 - np.sum((y_true - y_pred) ** 2.0) / variance

@guvectorize([(float64[:], float64[:], float64[:])], "(n),(n)->()")
def relative_mean_bias(
    y_true: npt.NDArray[np.float64],
    y_pred: npt.NDArray[np.float64],
    result: npt.NDArray[np.float64]
    ) -> None:
    """
    Numba and polars compatible implementation of signed mean relative bias.
    Also called, mean relative error or fractional bias.
        
    Parameters
    ----------
    y_true: NDArray[np.float64], required
        Ground truth (correct) target values, also called observations, measurements, or observed values.
    y_pred: NDArray[np.float64], required
        Estimated target values, also called simulations or modeled values.
    result: NDArray[np.float64], required
        Stores scalar result.
        
    Returns
    -------
    None
    """
    total = np.sum(y_true)
    if total == 0.0:
        result[0] = np.nan
        return
    result[0] = np.sum(y_pred - y_true) / np.sum(y_true)

@guvectorize([(float64[:], float64[:], float64[:])], "(n),(n)->()")
def pearson_correlation_coefficient(
    y_true: npt.NDArray[np.float64],
    y_pred: npt.NDArray[np.float64],
    result: npt.NDArray[np.float64]
    ) -> None:
    """
    Numba and polars compatible implementation of the Pearson correlation
    coefficient.
        
    Parameters
    ----------
    y_true: NDArray[np.float64], required
        Ground truth (correct) target values, also called observations, measurements, or observed values.
    y_pred: NDArray[np.float64], required
        Estimated target values, also called simulations or modeled values.
    result: NDArray[np.float64], required
        Stores scalar result.
        
    Returns
    -------
    None
    """
    y_true_dev = y_true - np.mean(y_true)
    y_pred_dev = y_pred - np.mean(y_pred)
    num = np.sum(y_true_dev * y_pred_dev)
    den = (
        np.sqrt(np.sum(y_true_dev ** 2)) *
        np.sqrt(np.sum(y_pred_dev ** 2))
        )
    if den == 0:
        result[0] = np.nan
        return
    result[0] = num / den

@guvectorize([(float64[:], float64[:], float64[:])], "(n),(n)->()")
def relative_variability(
    y_true: npt.NDArray[np.float64],
    y_pred: npt.NDArray[np.float64],
    result: npt.NDArray[np.float64]
    ) -> None:
    """
    Numba and polars compatible implementation of relative variability,
    required to compute Kling-Gupta Model Efficiency.
        
    Parameters
    ----------
    y_true: NDArray[np.float64], required
        Ground truth (correct) target values, also called observations, measurements, or observed values.
    y_pred: NDArray[np.float64], required
        Estimated target values, also called simulations or modeled values.
    result: NDArray[np.float64], required
        Stores scalar result.
        
    Returns
    -------
    None
    """
    std_dev = np.std(y_true)
    if std_dev == 0:
        result[0] = np.nan
        return
    result[0] = np.std(y_pred) / std_dev

@guvectorize([(float64[:], float64[:], float64[:])], "(n),(n)->()")
def relative_mean(
    y_true: npt.NDArray[np.float64],
    y_pred: npt.NDArray[np.float64],
    result: npt.NDArray[np.float64]
    ) -> None:
    """
    Numba and polars compatible implementation of relative mean,
    required to compute Kling-Gupta Model Efficiency.
        
    Parameters
    ----------
    y_true: NDArray[np.float64], required
        Ground truth (correct) target values, also called observations, measurements, or observed values.
    y_pred: NDArray[np.float64], required
        Estimated target values, also called simulations or modeled values.
    result: NDArray[np.float64], required
        Stores scalar result.
        
    Returns
    -------
    None
    """
    mean = np.mean(y_true)
    if mean == 0:
        result[0] = np.nan
        return
    result[0] = np.std(y_pred) / mean

@guvectorize([(float64[:], float64[:], float64[:])], "(n),(n)->()")
def kling_gupta_efficiency(
    y_true: npt.NDArray[np.float64],
    y_pred: npt.NDArray[np.float64],
    result: npt.NDArray[np.float64]
    ) -> None:
    """Returns an expression used to add a 'kling_gupta_efficiency' column to a dataframe."""
    correlation = np.array([0.0], dtype=np.float64)
    pearson_correlation_coefficient(y_true, y_pred, correlation)
    rel_var = np.array([0.0], dtype=np.float64)
    relative_variability(y_true, y_pred, rel_var)
    rel_mean = np.array([0.0], dtype=np.float64)
    relative_mean(y_true, y_pred, rel_mean)
    result[0] = (1.0 - np.sqrt(
        ((correlation[0] - 1.0)) ** 2.0 + 
        ((rel_var[0] - 1.0)) ** 2.0 + 
        ((rel_mean[0] - 1.0)) ** 2.0
        ))

METRIC_FUNCTIONS: dict[str, Callable[[npt.NDArray[np.float64], npt.NDArray[np.float64]], npt.NDArray[np.float64]]] = {
    "nse": nash_sutcliffe_efficiency,
    "rmb": relative_mean_bias,
    "pearson": pearson_correlation_coefficient,
    "rel_mean": relative_mean,
    "rel_var": relative_variability,
    "kge": kling_gupta_efficiency
}
"""Metrics to compute."""

SUFFIXES: list[str] = ["", "_lower", "_upper"]
"""Metric field suffixes."""

METRIC_FIELDS: list[str] = [k+s for k in METRIC_FUNCTIONS.keys() for s in SUFFIXES]
"""Metric field names."""

def compute_metrics(
    y_true: pl.Series,
    y_pred: pl.Series,
    minimum_sample_size: int = 30,
    minimum_mean: float = 0.01,
    minimum_variance: float = 0.000025
    ) -> npt.NDArray[np.float64]:
    results = []
    for _, func in METRIC_FUNCTIONS.items():
        # TODO check for nan?

        # Point estimates
        point_estimate = func(
            y_true.to_numpy(), 
            y_pred.to_numpy()
            )

        # Low sample size
        if len(y_true) < minimum_sample_size:
            results.append(np.asarray([point_estimate, np.nan, np.nan]))
            continue
        
        # Mostly zero values
        if y_true.mean() < minimum_mean:
            results.append(np.asarray([point_estimate, np.nan, np.nan]))
            continue
        
        # Nearly constant values
        if y_true.var() < minimum_variance:
            results.append(np.asarray([point_estimate, np.nan, np.nan]))
            continue
        
        # Optimal block size
        max_value = y_true.max() * 1.01
        normalized = y_true.to_numpy() / max_value
        block_size = optimal_block_length(normalized)["stationary"][0]
        if np.isnan(block_size):
            block_size = 1
        else:
            block_size = max(1, int(block_size))
        
        # Bootstrap confidence interval for each metric
        obs = y_true.to_numpy()
        pred = y_pred.to_numpy()
        indices = np.arange(obs.size)
        bs = StationaryBootstrap(
            block_size,
            indices,
            seed=2025
            )
        posterior = []
        for samples in bs.bootstrap(1000):
            # Certain functions may need additional error checks
            idx = samples[0][0]
            posterior.append(func(obs[idx], pred[idx]))
        ci = np.quantile(posterior, [0.025, 0.975])
        results.append(np.concatenate(([point_estimate], ci)))
    return np.concatenate(results)

def compute_metrics_pandas(
    data: pd.DataFrame,
    minimum_sample_size: int = 30,
    minimum_mean: float = 0.01,
    minimum_variance: float = 0.000025
    ) -> pd.Series:
    y_true = data["observed"].to_numpy(dtype=np.float64)
    y_pred = data["predicted"].to_numpy(dtype=np.float64)
    sample_size = data["observed"].count()

    results = []
    for _, func in METRIC_FUNCTIONS.items():
        # TODO check for nan?

        # Point estimates
        point_estimate = func(
            y_true, 
            y_pred
            )

        # Low sample size
        if sample_size < minimum_sample_size:
            results.append(np.asarray([point_estimate, np.nan, np.nan]))
            continue
        
        # Mostly zero values
        if np.mean(y_true) < minimum_mean:
            results.append(np.asarray([point_estimate, np.nan, np.nan]))
            continue
        
        # Nearly constant values
        if np.var(y_true) < minimum_variance:
            results.append(np.asarray([point_estimate, np.nan, np.nan]))
            continue
        
        # Optimal block size
        max_value = np.max(y_true) * 1.01
        normalized = y_true / max_value
        block_size = optimal_block_length(normalized)["stationary"][0]
        if np.isnan(block_size):
            block_size = 1
        else:
            block_size = max(1, int(block_size))
        
        # Bootstrap confidence interval for each metric
        indices = np.arange(y_true.size)
        bs = StationaryBootstrap(
            block_size,
            indices,
            seed=2025
            )
        posterior = []
        for samples in bs.bootstrap(1000):
            # Certain functions may need additional error checks
            idx = samples[0][0]
            posterior.append(func(y_true[idx], y_pred[idx]))
        ci = np.quantile(posterior, [0.025, 0.975])
        results.append(np.concatenate(([point_estimate], ci)))
    s = pd.Series(
        np.concatenate(results),
        index=METRIC_FIELDS
    )
    s["sample_size"] = sample_size
    s["start_date"] = data["value_time"].min()
    s["end_date"] = data["value_time"].max()
    s["nwm_feature_id"] = data["nwm_feature_id"].iloc[0]
    s["usgs_site_code"] = data["usgs_site_code"].iloc[0]
    if "lead_time_hours_min" in data:
        s["lead_time_hours_min"] = data["lead_time_hours_min"].min()
    return s



"""Plot managers."""
from dataclasses import dataclass
from typing import Any
import plotly.graph_objects as go
import colorcet as cc
import numpy.typing as npt

from nwm_explorer.mappings import DEFAULT_ZOOM, Domain
from nwm_explorer.readers import RoutelinkReader

SITE_MAP_CUSTOM_DATA_COLUMNS: list[str] = [
    "usgs_site_code",
    "nwm_feature_id"
]
"""Custom data column labels for use with Plotly hover tooltips."""

SITE_MAP_HOVER_TEMPLATE: str = (
    "USGS Site Code: %{customdata[0]}<br>"
    "NWM Feature ID: %{customdata[1]}<br>"
    "Longitude: %{lon}<br>"
    "Latitude: %{lat}"
)
"""Plotly compatible hover template for site maps."""

@dataclass
class SiteMapPlotter:
    scatter: go.Scattermap | None = None
    layout: go.Layout | None = None

    def __post_init__(self) -> None:
        if self.scatter is None:
            self.scatter = go.Scattermap(
                showlegend=False,
                name="",
                mode="markers",
                marker=dict(
                    size=15,
                    colorscale=cc.gouldian
                    )
                )
        if self.layout is None:
            self.layout = go.Layout(
                showlegend=False,
                height=720,
                width=1280,
                margin=dict(l=0, r=0, t=50, b=0),
                map=dict(
                    style="satellite-streets",
                ),
                clickmode="event",
                modebar=dict(
                    remove=["lasso", "select"]
                ),
                dragmode="zoom"
            )

        self.figure = {
            "data": [self.scatter],
            "layout": self.layout
        }

    def update_points(
            self,
            domain: Domain,
            values: npt.ArrayLike,
            metric_label: str,
            routelink_reader: RoutelinkReader
            ) -> tuple[float, float, float]:
        geometry = routelink_reader.geometry(domain)
        self.scatter.update(dict(
            lat=geometry.y,
            lon=geometry.x,
            customdata=routelink_reader.select_columns(
                domain,
                SITE_MAP_CUSTOM_DATA_COLUMNS
            ),
            hovertemplate=(
                SITE_MAP_HOVER_TEMPLATE +
                f"<br>{metric_label}: " +
                "%{marker.color:.2f}"
            )
        ))
        self.scatter["marker"].update(dict(
            color=values,
            colorbar=dict(
                title=dict(
                    text=metric_label,
                    side="right"
                    )
                ),
                cmin=-1.0,
                cmax=1.0
        ))
        lat = geometry.y.mean()
        lon = geometry.x.mean()
        zoom = DEFAULT_ZOOM[domain]
        self.layout["map"].update(dict(
            center={
                "lat": lat,
                "lon": lon
                },
            zoom=zoom
        ))
        return lat, lon, zoom

    def update_colors(
            self,
            values: npt.ArrayLike,
            metric_label: str,
            relayout_data: dict[str, Any]
            ) -> None:
        self.scatter.update(dict(
            hovertemplate=(
                SITE_MAP_HOVER_TEMPLATE +
                f"<br>{metric_label}: " +
                "%{marker.color:.2f}"
            )
        ))
        self.scatter["marker"].update(dict(
            color=values,
            colorbar=dict(
                title=dict(
                    text=metric_label,
                    side="right"
                    )
                ),
                cmin=-1.0,
                cmax=1.0
        ))
        if "map.center" in relayout_data:
            self.layout["map"]["center"].update({
                "lat": relayout_data["map.center"]["lat"],
                "lon": relayout_data["map.center"]["lon"]
            })
            self.layout["map"].update({
                "zoom": relayout_data["map.zoom"]
            })



"""Various standard procedures."""
from concurrent.futures import ProcessPoolExecutor
from pathlib import Path
import warnings
import pandas as pd
import polars as pl

from nwm_explorer.urls import generate_reference_dates, generate_usgs_urls
from nwm_explorer.manifests import generate_default_manifest, generate_usgs_manifest
from nwm_explorer.downloads import download_files, download_routelinks
from nwm_explorer.data import scan_routelinks, generate_filepath, generate_directory
from nwm_explorer.data import (process_netcdf_parallel, process_nwis_tsv_parallel,
    delete_directory)
from nwm_explorer.mappings import FileType, Variable, Units, Domain, Configuration
from nwm_explorer.mappings import LEAD_TIME_FREQUENCY, NWM_URL_BUILDERS
from nwm_explorer.metrics import compute_metrics_pandas
from nwm_explorer.data import netcdf_validator, csv_gz_validator
from nwm_explorer.logger import get_logger
from nwm_explorer.readers import read_pairs, read_NWM_output, read_USGS_observations, scan_date_range

def load_NWM_output(
        root: Path,
        start_date: pd.Timestamp,
        end_date: pd.Timestamp,
        routelinks: dict[Domain, pl.LazyFrame] | None = None,
        low_memory: bool = False
) -> dict[tuple[Domain, Configuration], pl.LazyFrame]:
    """
    Download and process NWM output.

    Parameters
    ----------
    root: Path, required
        Root directory to save downloaded and processed files.
    start_date: pd.Timestamp
        First date to start retrieving data.
    end_date: pd.Timestamp
        Last date to retrieve data.
    routelinks: dict[Domain, LazyFrame]
        Mapping from Domain to crosswalk data.
    low_memory: bool, default False
        Reduce memory pressure at the expense of performance.
    
    Returns
    -------
    dict[tuple[Domain, Configuration], pl.LazyFrame]
    """
    if routelinks is None:
        routelinks = scan_routelinks(*download_routelinks(root / "routelinks"))

    logger = get_logger("nwm_explorer.pipelines.load_NWM_output")
    reference_dates = generate_reference_dates(start_date, end_date)

    # Download and process model output
    model_output = {}
    for (domain, configuration), url_builder in NWM_URL_BUILDERS.items():
        day_files = []
        for rd in reference_dates:
            # Check for file existence
            parquet_file = generate_filepath(
                root, FileType.parquet, configuration, Variable.streamflow,
                Units.cubic_feet_per_second, rd
            )
            logger.info(f"Building {parquet_file}")
            if parquet_file.exists():
                logger.info(f"Found existing {parquet_file}")
                day_files.append(parquet_file)
                continue

            # Download
            urls = url_builder([rd])
            download_directory = generate_directory(
                root, FileType.netcdf, rd, configuration=configuration
            )
            logger.info(f"Downloading to {download_directory}")
            manifest = generate_default_manifest(len(urls),
                directory=download_directory)
            download_files(*zip(urls, manifest), limit=20, timeout=3600,
                file_validator=netcdf_validator)
            
            # Validate manifest
            logger.info(f"Validating manifest {domain} {configuration} {rd}")
            file_list = []
            for fp in manifest:
                if fp.exists():
                    file_list.append(fp)
                    continue
                warnings.warn(f"{fp} does not exist.", RuntimeWarning)
            
            # Check for at least one file
            if not file_list:
                logger.info(f"Found no files for {domain} {configuration} {rd}")
                continue

            # Process
            logger.info(f"Processing raw data {domain} {configuration} {rd}")
            features = routelinks[domain].select(
                "nwm_feature_id").collect()["nwm_feature_id"].to_list()
            data = process_netcdf_parallel(
                filepaths=file_list,
                variables=["streamflow"],
                features=features,
                max_processes=6,
                files_per_job=15
            ).rename(columns={
                    "time": "value_time",
                    "feature_id": "nwm_feature_id",
                    "streamflow": "value"
            })

            # Convert from cms to cfs
            data["value"] = data["value"].div(0.3048 ** 3.0)

            # Save to parquet
            logger.info(f"Saving {parquet_file}")
            pl.DataFrame(data).with_columns(
                pl.col("value").cast(pl.Float32)
            ).write_parquet(parquet_file)

            # Clean-up
            logger.info(f"Cleaning up {download_directory}")
            delete_directory(download_directory, parquet_file)
            day_files.append(parquet_file)
            
        # Check for at least one file
        if not day_files:
            logger.info(f"Found no data for {domain} {configuration}")
            continue
        
        # Merge files
        logger.info(f"Merging parquet files")
        model_output[(domain, configuration)] = pl.scan_parquet(day_files,
            low_memory=low_memory)
    return model_output

def load_USGS_observations(
        root: Path,
        start_date: pd.Timestamp,
        end_date: pd.Timestamp,
        routelinks: dict[Domain, pl.LazyFrame]
) -> dict[Domain, pl.LazyFrame]:
    """
    Download and process USGS observations.

    Parameters
    ----------
    root: Path, required
        Root directory to save downloaded and processed files.
    start_date: pd.Timestamp
        First date to start retrieving data.
    end_date: pd.Timestamp
        Last date to retrieve data.
    routelinks: dict[Domain, LazyFrame]
        Mapping from Domain to crosswalk data.
    
    Returns
    -------
    dict[Domain, pl.LazyFrame]
    """
    logger = get_logger("nwm_explorer.pipelines.load_USGS_observations")
    # Download and process model output
    observations = {}
    s = start_date.strftime("%Y%m%dT%H")
    e = end_date.strftime("%Y%m%dT%H")
    for domain, rl in routelinks.items():
        # Check for file existence
        parquet_directory = root / FileType.parquet / Configuration.usgs
        parquet_file = parquet_directory / f"{domain}_{s}_{e}.parquet"
        logger.info(f"Building {parquet_file}")
        if parquet_file.exists():
            logger.info(f"Found existing {parquet_file}")
            observations[domain] = pl.scan_parquet(parquet_file)
            continue
        parquet_directory.mkdir(exist_ok=True, parents=True)

        # Download
        sites = rl.select(
            "usgs_site_code").collect().to_pandas()["usgs_site_code"]
        sites = sites[sites.str.isdigit()].to_list()
        urls = generate_usgs_urls(
            sites, start_date, end_date
        )
        download_directory = root / FileType.tsv / domain
        download_directory.mkdir(exist_ok=True, parents=True)
        logger.info(f"Downloading to {download_directory}")
        manifest = generate_usgs_manifest(
            sites,
            directory=download_directory)
        download_files(*zip(urls, manifest), limit=10, timeout=3600, 
            headers={"Accept-Encoding": "gzip"}, auto_decompress=False,
            file_validator=csv_gz_validator)
        
        # Validate manifest
        logger.info(f"Validating manifest {domain}")
        file_list = []
        for fp in manifest:
            if fp.exists():
                file_list.append(fp)
                continue
            warnings.warn(f"{fp} does not exist.", RuntimeWarning)

        # Process
        logger.info(f"Processing raw data {domain}")
        data = process_nwis_tsv_parallel(
            filepaths=file_list,
            max_processes=12
        ).rename(columns={
                "time": "value_time",
                "feature_id": "nwm_feature_id",
                "streamflow": "value"
        })

        # Save to parquet
        logger.info(f"Saving {parquet_file}")
        pl.DataFrame(data).write_parquet(parquet_file)
        observations[domain] = pl.scan_parquet(parquet_file)

        # Clean-up
        logger.info(f"Cleaning up {download_directory}")
        delete_directory(download_directory, parquet_file)
    return observations

def load_pairs(
        root: Path,
        start_date: pd.Timestamp,
        end_date: pd.Timestamp
) -> dict[tuple[Domain, Configuration], pl.LazyFrame]:
    """
    Download and process NWM and USGS output.

    Parameters
    ----------
    root: Path, required
        Root directory to save downloaded and processed files.
    start_date: pd.Timestamp
        First date to start retrieving data.
    end_date: pd.Timestamp
        Last date to retrieve data.
    
    Returns
    -------
    dict[tuple[Domain, Configuration], pl.LazyFrame]
    """
    logger = get_logger("nwm_explorer.pipelines.load_pairs")
    routelinks = scan_routelinks(*download_routelinks(root / "routelinks"))

    logger.info("Scanning predictions for valid time range")
    predictions = read_NWM_output(
        root=root,
        start_date=start_date,
        end_date=end_date,
        low_memory=True
    )
    first, last = scan_date_range(predictions)
    logger.info("Loading observations")
    observations = read_USGS_observations(
        root=root,
        start_date=first,
        end_date=last
    )

    # Resample obs
    logger.info("Resampling observations")
    resampled_obs = {}
    for (domain, configuration), _ in predictions.items():
        if configuration in LEAD_TIME_FREQUENCY:
            resampling_frequency = LEAD_TIME_FREQUENCY[configuration][1]
            if (domain, resampling_frequency) in resampled_obs:
                continue
            resampled_obs[(
                domain,
                resampling_frequency
                )] = observations[domain].sort(
                    ("usgs_site_code", "value_time")
                ).group_by_dynamic(
                    "value_time",
                    every=resampling_frequency,
                    group_by="usgs_site_code"
                ).agg(
                    pl.col("value").max().alias("observed")
                ).with_columns(
                    pl.col("usgs_site_code").cast(pl.String)
                ).collect()
            continue
    
    for domain in observations:
        if (domain, "1d") in resampled_obs:
            continue
        resampled_obs[(
            domain,
            "1d"
            )] = observations[domain].sort(
                ("usgs_site_code", "value_time")
            ).group_by_dynamic(
                "value_time",
                every="1d",
                group_by="usgs_site_code"
            ).agg(
                pl.col("value").max().alias("observed")
            ).with_columns(
                pl.col("usgs_site_code").cast(pl.String)
            ).collect()
        
    # Load crosswalks
    logger.info("Loading crosswalks")
    crosswalks = {}
    for domain in routelinks:
        crosswalks[domain] = routelinks[domain].select(["nwm_feature_id",
                "usgs_site_code"]).collect()

    # Process model output
    logger.info("Resampling predictions and pairing")
    reference_dates = generate_reference_dates(start_date, end_date)
    pairs = {}
    for (domain, configuration), _ in predictions.items():
        day_files = []
        xwalk = crosswalks[domain]
        for rd in reference_dates:
            # Check for file existence
            parquet_file = generate_filepath(
                root, FileType.parquet, configuration, Variable.streamflow_pairs,
                Units.cubic_feet_per_second, rd
            )
            if parquet_file.exists():
                logger.info(f"Found existing file: {parquet_file}")
                day_files.append(parquet_file)
                # parquet_file.unlink()
                continue

            # Load input data
            ifile = generate_filepath(
                root, FileType.parquet, configuration, Variable.streamflow,
                Units.cubic_feet_per_second, rd
            )
            if not ifile.exists():
                logger.info(f"File does not exist, skipping: {ifile}")
                continue
            data = pl.scan_parquet(ifile).collect()
            logger.info(f"Building {parquet_file}")
            if configuration in LEAD_TIME_FREQUENCY:
                sampling_duration = LEAD_TIME_FREQUENCY[configuration][0]
                resampling_frequency = LEAD_TIME_FREQUENCY[configuration][1]
                hours = sampling_duration / pl.duration(hours=1)
                obs = resampled_obs[(domain, resampling_frequency)]
                paired_data = data.sort(
                    ("nwm_feature_id", "reference_time", "value_time")
                ).with_columns(
                    ((pl.col("value_time").sub(
                        pl.col("reference_time")
                        ) / sampling_duration).floor() *
                            hours).alias("lead_time_hours_min")
                ).group_by_dynamic(
                    "value_time",
                    every=resampling_frequency,
                    group_by=("nwm_feature_id", "reference_time")
                ).agg(
                    pl.col("value").max().alias("predicted"),
                    pl.col("lead_time_hours_min").min()
                ).with_columns(
                    usgs_site_code=pl.col("nwm_feature_id").replace_strict(
                    xwalk["nwm_feature_id"], xwalk["usgs_site_code"])
                ).join(obs, on=["usgs_site_code", "value_time"], how="left"
                    ).drop_nulls()
            else:
                # NOTE This will result in two simulation values per
                #  reference day. Handle this before computing metrics (max).
                obs = resampled_obs[(domain, "1d")]
                paired_data = data.sort(
                    ("nwm_feature_id", "value_time")
                ).group_by_dynamic(
                    "value_time",
                    every="1d",
                    group_by="nwm_feature_id"
                ).agg(
                    pl.col("value").max().alias("predicted")
                ).with_columns(
                usgs_site_code=pl.col("nwm_feature_id").replace_strict(
                    xwalk["nwm_feature_id"], xwalk["usgs_site_code"])
                ).join(obs, on=["usgs_site_code", "value_time"], how="left"
                    ).drop_nulls()

            # Save to parquet
            logger.info(f"Saving {parquet_file}")
            paired_data.write_parquet(parquet_file)

            # Add file to list
            day_files.append(parquet_file)
            
        # Check for at least one file
        if not day_files:
            logger.info(f"Found no data for {domain} {configuration}")
            continue
        
        # Merge files
        logger.info(f"Merging parquet files")
        pairs[(domain, configuration)] = pl.scan_parquet(day_files)
    return pairs

def load_metrics(
        root: Path,
        start_date: pd.Timestamp,
        end_date: pd.Timestamp
) -> dict[tuple[Domain, Configuration], pl.LazyFrame]:
    """
    Download and process NWM and USGS output.

    Parameters
    ----------
    root: Path, required
        Root directory to save downloaded and processed files.
    start_date: pd.Timestamp
        First date to start retrieving data.
    end_date: pd.Timestamp
        Last date to retrieve data.
    
    Returns
    -------
    dict[tuple[Domain, Configuration], pl.LazyFrame]
    """
    logger = get_logger("nwm_explorer.pipelines.load_metrics")
    logger.info("Reading pairs")
    pairs = read_pairs(
        root=root,
        start_date=start_date,
        end_date=end_date
        )

    results = {}
    s = start_date.strftime("%Y%m%dT%H")
    e = end_date.strftime("%Y%m%dT%H")
    pool = ProcessPoolExecutor(max_workers=18)
    for (domain, configuration), paired in pairs.items():
        # Check for file existence
        parquet_directory = root / FileType.parquet / "metrics"
        parquet_file = parquet_directory / f"{domain}_{configuration}_{s}_{e}.parquet"
        logger.info(f"Building {parquet_file}")
        if parquet_file.exists():
            logger.info(f"Found existing {parquet_file}")
            results[(domain, configuration)] = pl.scan_parquet(parquet_file)
            continue
        parquet_directory.mkdir(exist_ok=True, parents=True)

        logger.info(f"Reading pairs {domain} {configuration}")
        if configuration in LEAD_TIME_FREQUENCY:
            groups = ["usgs_site_code", "lead_time_hours_min"]
            data = paired.sort(
                    ("nwm_feature_id", "reference_time", "value_time")
                ).group_by_dynamic(
                    "value_time",
                    every="1d",
                    group_by=["nwm_feature_id", "reference_time"]
                ).agg(
                    pl.col("observed").max(),
                    pl.col("predicted").max(),
                    pl.col("usgs_site_code").first(),
                    pl.col("lead_time_hours_min").min()
                ).with_columns(
                    pl.col("observed").cast(pl.Float32),
                    pl.col("usgs_site_code").cast(pl.Categorical),
                    pl.col("lead_time_hours_min").cast(pl.Int32)
                ).collect()
        else:
            groups = ["usgs_site_code"]
            data = paired.sort(
                    ("nwm_feature_id", "value_time")
                ).group_by_dynamic(
                    "value_time",
                    every="1d",
                    group_by="nwm_feature_id"
                ).agg(
                    pl.col("observed").max(),
                    pl.col("predicted").max(),
                    pl.col("usgs_site_code").first()
                ).with_columns(
                    pl.col("observed").cast(pl.Float32),
                    pl.col("usgs_site_code").cast(pl.Categorical)
                ).collect()

        logger.info(f"Computing metrics {domain} {configuration}")
        df = data.to_pandas()
        chunks = []
        for grp, chunk in df.groupby(groups, observed=True):
            chunks.append(chunk)
        chunk_size = max(1, int(len(chunks) / 18))
        jobs = list(pool.map(compute_metrics_pandas, chunks, chunksize=chunk_size))
        metric_results = pd.DataFrame(jobs)

        # Write results
        logger.info(f"Saving {parquet_file}")
        pl.DataFrame(metric_results).write_parquet(parquet_file)
        results[(domain, configuration)] = pl.scan_parquet(parquet_file)
    pool.shutdown()
    return results



"""Various mappings."""
from enum import StrEnum
from typing import Callable
import polars as pl
import pandas as pd
import nwm_explorer.urls as nwm_urls

EVALUATIONS: dict[str, tuple[pd.Timestamp, pd.Timestamp]] = {
    "FY2024_Q1": (pd.Timestamp("2023-10-01"), pd.Timestamp("2024-01-01")),
    "FY2024_Q2": (pd.Timestamp("2024-01-01"), pd.Timestamp("2024-04-01")),
    "FY2024_Q3": (pd.Timestamp("2024-04-01"), pd.Timestamp("2024-07-01")),
    "FY2024_Q4": (pd.Timestamp("2024-07-01"), pd.Timestamp("2024-10-01")),
    "FY2025_Q1": (pd.Timestamp("2024-10-01"), pd.Timestamp("2025-01-01")),
    "FY2025_Q2": (pd.Timestamp("2025-01-01"), pd.Timestamp("2025-04-01")),
    "FY2025_Q3": (pd.Timestamp("2025-04-01"), pd.Timestamp("2025-07-01"))
}
"""Standard evaluations."""

TIMEZONE_MAPPING: dict[str, str] = {
    "AKST": "America/Anchorage",
    "AKDT": "America/Anchorage",
    "HST": "America/Adak",
    "HDT": "America/Adak",
    "AST": "America/Puerto_Rico",
    "CDT": "America/Chicago",
    "CST": "America/Chicago",
    "EDT": "America/New_York",
    "EST": "America/New_York",
    "MST": "America/Phoenix",
    "MDT": "America/Denver",
    "PST": "America/Los_Angeles",
    "PDT": "America/Los_Angeles"
}
"""Mapping from common timezone strings to IANA compatible strings."""

ROUTELINK_SCHEMA: dict[str, pl.DataType] = {
    "nwm_feature_id": pl.Int64,
    "usgs_site_code": pl.String,
    "latitude": pl.Float64,
    "longitude": pl.Float64
}
"""Mapping from routelink field to polars datatype."""

class Domain(StrEnum):
    """Symbols used to reference different model domains."""
    alaska = "alaska"
    conus = "conus"
    hawaii = "hawaii"
    puertorico = "puertorico"

DOMAIN_MAPPING: dict[str, Domain] = {
    "alaska": Domain.alaska,
    "conus": Domain.conus,
    "hawaii": Domain.hawaii,
    "puertorico": Domain.puertorico,
    "RouteLink_AK.csv": Domain.alaska,
    "RouteLink_CONUS.csv": Domain.conus,
    "RouteLink_HI.csv": Domain.hawaii,
    "RouteLink_PRVI.csv": Domain.puertorico,
    "Alaska": Domain.alaska,
    "CONUS": Domain.conus,
    "Hawaii": Domain.hawaii,
    "Puerto Rico": Domain.puertorico
}
"""Mapping from common strings to standard symbols."""

class Configuration(StrEnum):
    """Symbols used to reference data configurations."""
    analysis_assim_extend_alaska_no_da = "analysis_assim_extend_alaska_no_da"
    analysis_assim_extend_no_da = "analysis_assim_extend_no_da"
    analysis_assim_hawaii_no_da = "analysis_assim_hawaii_no_da"
    analysis_assim_puertorico_no_da = "analysis_assim_puertorico_no_da"
    medium_range_mem1 = "medium_range_mem1"
    medium_range_blend = "medium_range_blend"
    medium_range_no_da = "medium_range_no_da"
    medium_range_alaska_mem1 = "medium_range_alaska_mem1"
    medium_range_blend_alaska = "medium_range_blend_alaska"
    medium_range_alaska_no_da = "medium_range_alaska_no_da"
    short_range = "short_range"
    short_range_alaska = "short_range_alaska"
    short_range_hawaii = "short_range_hawaii"
    short_range_hawaii_no_da = "short_range_hawaii_no_da"
    short_range_puertorico = "short_range_puertorico"
    short_range_puertorico_no_da = "short_range_puertorico_no_da"
    usgs = "usgs"

LEAD_TIME_FREQUENCY: dict[Configuration, pl.Duration] = {
    Configuration.medium_range_mem1: (pl.duration(hours=24), "1d"),
    Configuration.medium_range_blend: (pl.duration(hours=24), "1d"),
    Configuration.medium_range_no_da: (pl.duration(hours=24), "1d"),
    Configuration.medium_range_alaska_mem1: (pl.duration(hours=24), "1d"),
    Configuration.medium_range_blend_alaska: (pl.duration(hours=24), "1d"),
    Configuration.medium_range_alaska_no_da: (pl.duration(hours=24), "1d"),
    Configuration.short_range: (pl.duration(hours=6), "6h"),
    Configuration.short_range_alaska: (pl.duration(hours=5), "5h"),
    Configuration.short_range_hawaii: (pl.duration(hours=6), "6h"),
    Configuration.short_range_hawaii_no_da: (pl.duration(hours=6), "6h"),
    Configuration.short_range_puertorico: (pl.duration(hours=6), "6h"),
    Configuration.short_range_puertorico_no_da: (pl.duration(hours=6), "6h")
}
"""Mapping used for computing lead time and sampling frequency."""

class FileType(StrEnum):
    """Symbols used for common file types."""
    netcdf = "netcdf"
    parquet = "parquet"
    tsv = "tsv"

class Variable(StrEnum):
    """Symbols used for common variables."""
    streamflow = "streamflow"
    streamflow_pairs = "streamflow_pairs"
    streamflow_metrics = "streamflow_metrics"

class Units(StrEnum):
    """Symbols used for common units."""
    cubic_feet_per_second = "cfs"
    metrics = "metrics"

class Confidence(StrEnum):
    """Symbols used to describe confidence interval range estimates."""
    point = "point"
    lower = "lower"
    upper = "upper"

DEFAULT_ZOOM: dict[Domain, int] = {
    Domain.alaska: 5,
    Domain.conus: 3,
    Domain.hawaii: 6,
    Domain.puertorico: 8
}
"""Default map zoom for each domain."""

NWM_URL_BUILDERS: dict[tuple[Domain, Configuration], Callable] = {
    (Domain.alaska, Configuration.analysis_assim_extend_alaska_no_da): nwm_urls.analysis_assim_extend_alaska_no_da,
    (Domain.conus, Configuration.analysis_assim_extend_no_da): nwm_urls.analysis_assim_extend_no_da,
    (Domain.hawaii, Configuration.analysis_assim_hawaii_no_da): nwm_urls.analysis_assim_hawaii_no_da,
    (Domain.puertorico, Configuration.analysis_assim_puertorico_no_da): nwm_urls.analysis_assim_puertorico_no_da,
    (Domain.conus, Configuration.medium_range_mem1): nwm_urls.medium_range_mem1,
    (Domain.conus, Configuration.medium_range_blend): nwm_urls.medium_range_blend,
    (Domain.conus, Configuration.medium_range_no_da): nwm_urls.medium_range_no_da,
    (Domain.alaska, Configuration.medium_range_alaska_mem1): nwm_urls.medium_range_alaska_mem1,
    (Domain.alaska, Configuration.medium_range_blend_alaska): nwm_urls.medium_range_blend_alaska,
    (Domain.alaska, Configuration.medium_range_alaska_no_da): nwm_urls.medium_range_alaska_no_da,
    (Domain.conus, Configuration.short_range): nwm_urls.short_range,
    (Domain.alaska, Configuration.short_range_alaska): nwm_urls.short_range_alaska,
    (Domain.hawaii, Configuration.short_range_hawaii): nwm_urls.short_range_hawaii,
    (Domain.hawaii, Configuration.short_range_hawaii_no_da): nwm_urls.short_range_hawaii_no_da,
    (Domain.puertorico, Configuration.short_range_puertorico): nwm_urls.short_range_puertorico,
    (Domain.puertorico, Configuration.short_range_puertorico_no_da): nwm_urls.short_range_puertorico_no_da
}
"""Mapping from (Domain, Configuration) to url builder function."""

DOMAIN_CONFIGURATION_MAPPING: dict[Domain, dict[str, Configuration]] = {
    Domain.alaska: {
        "Extended Analysis (MRMS, No-DA)": Configuration.analysis_assim_extend_alaska_no_da,
        "Medium Range Forecast (GFS, Deterministic)": Configuration.medium_range_alaska_mem1,
        "Medium Range Forecast (NBM, Deterministic)": Configuration.medium_range_blend_alaska,
        "Medium Range Forecast (GFS, Deterministic, No-DA)": Configuration.medium_range_alaska_no_da,
        "Short Range Forecast (HRRR)": Configuration.short_range_alaska
    },
    Domain.conus: {
        "Extended Analysis (MRMS, No-DA)": Configuration.analysis_assim_extend_no_da,
        "Medium Range Forecast (GFS, Deterministic)": Configuration.medium_range_mem1,
        "Medium Range Forecast (NBM, Deterministic)": Configuration.medium_range_blend,
        "Medium Range Forecast (GFS, Deterministic, No-DA)": Configuration.medium_range_no_da,
        "Short Range Forecast (HRRR)": Configuration.short_range
    },
    Domain.hawaii: {
        "Analysis (MRMS, No-DA)": Configuration.analysis_assim_hawaii_no_da,
        "Short Range Forecast (WRF-ARW)": Configuration.short_range_hawaii,
        "Short Range Forecast (WRF-ARW, No-DA)": Configuration.short_range_hawaii_no_da
    },
    Domain.puertorico: {
        "Analysis (MRMS, No-DA)": Configuration.analysis_assim_puertorico_no_da,
        "Short Range Forecast (WRF-ARW)": Configuration.short_range_puertorico,
        "Short Range Forecast (WRF-ARW, No-DA)": Configuration.short_range_puertorico_no_da
    }
}
"""Mapping from domains to pretty string representations of model configurations."""

LEAD_TIME_VALUES: dict[Configuration, list[int]] = {
    Configuration.medium_range_mem1: [l for l in range(0, 240, 24)],
    Configuration.medium_range_blend: [l for l in range(0, 240, 24)],
    Configuration.medium_range_no_da: [l for l in range(0, 240, 24)],
    Configuration.medium_range_alaska_mem1: [l for l in range(0, 240, 24)],
    Configuration.medium_range_blend_alaska: [l for l in range(0, 240, 24)],
    Configuration.medium_range_alaska_no_da: [l for l in range(0, 240, 24)],
    Configuration.short_range: [l for l in range(0, 18, 6)],
    Configuration.short_range_alaska: [l for l in range(0, 45, 5)],
    Configuration.short_range_hawaii: [l for l in range(0, 48, 6)],
    Configuration.short_range_hawaii_no_da: [l for l in range(0, 48, 6)],
    Configuration.short_range_puertorico: [l for l in range(0, 48, 6)],
    Configuration.short_range_puertorico_no_da: [l for l in range(0, 48, 6)]
}
"""Mapping used to display available lead times."""

DOMAIN_STRINGS: dict[str, Domain] = {
    "Alaska": DOMAIN_MAPPING["Alaska"],
    "CONUS": DOMAIN_MAPPING["CONUS"],
    "Hawaii": DOMAIN_MAPPING["Hawaii"],
    "Puerto Rico": DOMAIN_MAPPING["Puerto Rico"]
}
"""Mapping from pretty strings to domains for display."""

CONFIDENCE_STRINGS: dict[str, Confidence] = {
    "Point": Confidence.point,
    "Lower": Confidence.lower,
    "Upper": Confidence.upper
}
"""Mapping from pretty strings to confidence boundaries for display."""

class Metric(StrEnum):
    """Symbols used to reference different evaluation metrics."""
    nash_sutcliffe_efficiency = "nash_sutcliffe_efficiency"
    mean_relative_bias = "mean_relative_bias"
    pearson_correlation_coefficient = "pearson_correlation_coefficient"
    relative_variability = "relative_variability"
    relative_mean = "relative_mean"
    kling_gupta_efficiency = "kling_gupta_efficiency"

METRIC_STRINGS: dict[str, Metric] = {
    "Mean relative bias": Metric.mean_relative_bias,
    "Pearson correlation coefficient": Metric.pearson_correlation_coefficient,
    "Nash-Sutcliffe efficiency": Metric.nash_sutcliffe_efficiency,
    "Relative mean": Metric.relative_mean,
    "Relative variability": Metric.relative_variability,
    "Kling-Gupta efficiency": Metric.kling_gupta_efficiency
}
"""Mapping from pretty strings to evaluation metrics for display."""
